import argparse
import os

from bal_xilinx.analyzers.device_analyzer import XilinxDeviceAnalyzer
from bal_xilinx.context import XilinxContextFactory
from bal_xilinx.defaults import default_xilinx_context, default_xilinx_formats
from bal_xilinx.format import XilinxFormatBuilder
from bal_xilinx.modifiers.pin_modifier import XilinxPinModifer


def update_xilinx_pin(bitstream_path, output_bitstream_path, is_on, pins):
    xilinx_context_factory = default_xilinx_context(XilinxContextFactory(
        default_xilinx_formats(XilinxFormatBuilder()).build()
    ))
    with open(bitstream_path, "rb") as f:
        data = f.read()

    bitstream_context = xilinx_context_factory.create(data)
    bitstream_object = bitstream_context.get_data()

    device_type = bitstream_context.create_analyzer(XilinxDeviceAnalyzer) \
        .analyze()
    if device_type != "LX9":
        print("The script only supports bitstreams targetting the LX9.")

    print("Modifying the bitstream")
    pin_modifier = bitstream_context.create_modifier(XilinxPinModifer)
    for pin in pins:
        pin_modifier.modify(pin, is_on)

    bitstream_object.synchronize(True)
    packed_data = bitstream_object.pack()
    assert packed_data != data, "The bitstream did not change."
    print("Writing modified bitstream to {}".format(output_bitstream_path))
    with open(output_bitstream_path, "wb") as f:
        f.write(packed_data)
    print("Done!")


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        prog="bal_xilinx.tools.pin",
        description='Modify a Xilinx FPGA bitstream to turn a specific pin on or off. It currently'
                    'supports bitstream targeting the LX9.'
    )
    parser.add_argument(
        'path',
        metavar='PATH',
        help='The path to the bitstream'
    )
    parser.add_argument(
        'state',
        metavar='STATE',
        choices=['on', 'off'],
        help='Determine if the pin should pull low (on) or high (off)'
    )
    parser.add_argument(
        'pins',
        metavar='PIN',
        nargs="+",
        help='The name of the pin to modify (ie P134)'
    )

    args = parser.parse_args()
    bitstream_output_path = os.path.splitext(args.path)[0] + "_repacked.bin"
    update_xilinx_pin(args.path, bitstream_output_path, args.state == 'on', args.pins)


from collections import OrderedDict

from bal.context_ioc import AbstractConverter
from bal.data_object import DataObject
from bal_xilinx.data_model import XilinxType1Payload, XilinxType1PayloadAttribute
from bal_xilinx.format import XilinxRegisterFormat


def pad_bytes(data, alignment=4):
    if len(data) % alignment == 0:
        return data
    data = bytearray(data)
    while len(data) % alignment != 0:
        data.append(0)
    return bytes(data)


class XilinxType1PayloadConverter(AbstractConverter):
    def __init__(self, context, register_format):
        """
        :param XilinxContext context: A factory used to create data objects/converters.
        :param XilinxRegisterFormat register_format: The format of the register being parsed.
        """
        super(XilinxType1PayloadConverter, self).__init__(context)
        self.context = context
        self.register_format = register_format

    def unpack(self, data_bytes):
        """
        :param bytes data_bytes:
        :rtype: XilinxType1Payload
        """
        if len(data_bytes) != self.register_format.size:
            raise ValueError(
                "Error while parsing the config for register {}. "
                "The expected size of the register config data ({}) "
                "does not match the size of the data to parse ({})".format(
                    self.register_format.name,
                    self.register_format.size,
                    len(data_bytes)
                )
            )
        data_bytes = pad_bytes(data_bytes)
        # Parse the register config attributes using the ctype defined by the
        attributes_values = self.register_format.ctype.from_buffer_copy(data_bytes)
        attributes = OrderedDict()
        for i, attribute_format in enumerate(self.register_format.attributes):
            value = attributes_values.values[i]
            value_documentation = attribute_format.get_value_documentation(value)
            if value_documentation is not None:
                value_name = value_documentation.name
                value_description = value_documentation.description
            else:
                value_name = None
                value_description = None
            attribute_name = attribute_format.name.lower()
            class_name = "".join([p.title() for p in attribute_name.split("_")]) + "Value"
            attributes[attribute_name] = self._create_attribute_object(
                attribute_format,
                class_name,
                value,
                value_name,
                value_description,
            )
        class_name = "Xilinx{}Payload".format(self.register_format.name)
        PayloadModel = type(class_name, (XilinxType1Payload,), {})
        PayloadModel.__doc__ = self.register_format.description
        return PayloadModel(attributes)

    def _create_attribute_object(
            self,
            attribute_format,
            class_name,
            value,
            value_name,
            value_description,
    ):
        AttributeModel = type(str(class_name), (XilinxType1PayloadAttribute, ), {})
        AttributeModel.__doc__ = attribute_format.description
        model = AttributeModel(int(value), value_name, value_description)
        return DataObject.create_unpacked(self.context, model, bit_size=attribute_format.bit_size)

    def pack(self, data_model):
        """
        :param XilinxType1Payload data_model:
        :rtype: bytes
        """
        assert isinstance(data_model, XilinxType1Payload)
        attributes = []
        for attribute_format in self.register_format.attributes:
            attributes.append(data_model.get(attribute_format.name.lower()).get_model().get_value())
        raw_attributes = self.register_format.ctype.get_bytes(*attributes)
        raw_attributes = raw_attributes[:self.register_format.size]
        return raw_attributes



import io

from typing import List

from bal.context_ioc import AbstractConverter
from bal.data_model import ArrayModel
from bal.data_object import DataObject
from bal_xilinx.context import XilinxContext
from bal_xilinx.data_model import XilinxFdriPayload, XilinxFdriLogicBlock, \
    XilinxFdriRAMBlockInterface, XilinxFdriIOBlockInterface, XilinxFdriCRCInterface, \
    XilinxFdriLogicRow, XilinxFdriLogicMajor, XilinxFdriLogicFrame
from bal_xilinx.format import XilinxFdriMajorFormat


class XilinxFdriLogicMajorConverter(AbstractConverter):
    def __init__(self, context, major_format):
        """
        :param XilinxContext context:
        :param XilinxFdriMajorFormat major_format: The format of the register being parsed.
        """
        super(XilinxFdriLogicMajorConverter, self).__init__(context)
        self.context = context
        self._major_format = major_format

    def unpack(self, data_bytes):
        """
        :param bytes data_bytes:
        :rtype: ArrayModel
        """
        frames = []
        data_stream = io.BytesIO(data_bytes)
        expected_size = 0
        for frame_index in range(self._major_format.frame_count):
            frame_data = data_stream.read(self._major_format.frame_size)
            expected_size += self._major_format.frame_size
            frame_description = None
            if len(self._major_format.frame_descriptions) > frame_index:
                frame_description = self._major_format.frame_descriptions[frame_index]
            FrameInterface = type(
                "XilinxFdriLogicFrame",
                (XilinxFdriLogicFrame,),
                {}
            )
            FrameInterface.__doc__ = frame_description
            frames.append(
                DataObject.create_packed(
                    self.context,
                    frame_data,
                    FrameInterface
                )
            )

        assert expected_size == len(data_bytes), \
            "The number of bytes provided {} does not match the expected number of bytes {}".format(
                len(data_bytes),
                expected_size

        )
        major_name = "".join([p.title() for p in self._major_format.name.split("_")])
        class_name = "{}".format(major_name)
        MajorModel = type(class_name, (XilinxFdriLogicMajor,), {})
        MajorModel.__doc__ = "The config data for a {} major".format(self._major_format.name)
        return MajorModel(frames)

    def pack(self, data_model):
        """
        :param ArrayModel data_model:
        :rtype: bytes
        """
        return b"".join([c.pack() for c in data_model])


class XilinxFdriLogicBlockRowConverter(AbstractConverter):
    """
    :param XilinxContext context:
    :param List[XilinxFdriMajorFormat] row_format: The format of the register being parsed.
    """
    def __init__(self, context, row_format):
        super(XilinxFdriLogicBlockRowConverter, self).__init__(context)
        self.context = context
        self._row_format = row_format

    def unpack(self, data_bytes):
        """
        :param bytes data_bytes:
        :rtype: ArrayModel
        """
        majors = []
        data_stream = io.BytesIO(data_bytes)
        expected_size = 0
        for major_format in self._row_format:
            major_size = major_format.frame_count * major_format.frame_size
            expected_size += major_size
            majors.append(DataObject.create_packed(
                self.context,
                data_stream.read(major_size),
                XilinxFdriLogicMajor,
                converter_args=(major_format, )
            ))

        assert expected_size == len(data_bytes), \
            "The number of bytes provided {} does not match the expected number of bytes {}".format(
                len(data_bytes),
                expected_size
            )
        return XilinxFdriLogicRow(majors)

    def pack(self, data_model):
        """
        :param ArrayModel data_model:
        :rtype: bytes
        """
        return b"".join([c.pack() for c in data_model])


class XilinxFdriLogicConverter(AbstractConverter):
    """
    :param XilinxContext context: A factory used to create data objects/converters.
    """
    def __init__(self, context):
        super(XilinxFdriLogicConverter, self).__init__(context)
        self.context = context

    def unpack(self, data_bytes):
        """
        :param bytes data_bytes:
        :rtype: ArrayModel
        """
        if self.context.id_code is None:
            raise ValueError(
                "The ID code for the targeted device has not been set on the bitstream context."
            )

        rows = []
        expected_size = 0
        data_stream = io.BytesIO(data_bytes)
        fdri_format = self.context.format.get_fdri_format(self.context.id_code)
        for row_format in fdri_format.logic_block_format:
            row_size = sum([
                major_format.frame_size * major_format.frame_count
                for major_format in row_format
            ])
            expected_size += row_size
            rows.append(DataObject.create_packed(
                self.context,
                data_stream.read(row_size),
                XilinxFdriLogicRow,
                converter_args=(row_format, ),
            ))

        assert expected_size == len(data_bytes), \
            "The number of bytes provided {} does not match the expected number of bytes {}".format(
                len(data_bytes),
                expected_size
            )
        return XilinxFdriLogicBlock(rows)

    def pack(self, data_object):
        """
        :param ArrayModel data_object:
        :rtype: bytes
        """
        return b"".join([c.pack() for c in data_object])


class XilinxFdriRAMBlockConverter(AbstractConverter):
    pass


class XilinxFdriIOBlockConverter(AbstractConverter):
    pass


class XilinxFdriTailConverter(AbstractConverter):
    pass


class XilinxFdriPayloadConverter(AbstractConverter):
    """
    :param XilinxContext context: A factory used to create data objects/converters.
    """
    def __init__(self, context):
        super(XilinxFdriPayloadConverter, self).__init__(context)
        self.context = context

    def unpack(self, data_bytes):
        """
        :param bytes data_bytes:
        :rtype: XilinxFdriPayload
        """
        if self.context.id_code is None:
            raise ValueError(
                "The ID code for the targeted device has not been set on the bitstream context."
            )
        fdri_format = self.context.format.get_fdri_format(self.context.id_code)
        expected_size = fdri_format.logic_block_size + fdri_format.bram_block_size + \
                fdri_format.io_block_size + fdri_format.crc_size
        assert len(data_bytes) == expected_size, \
            "The size of the config data ({}) does not match the expected size ({})".format(
                len(data_bytes),
                expected_size
            )
        data_stream = io.BytesIO(data_bytes)
        return XilinxFdriPayload(
            DataObject.create_packed(
                self.context,
                data_stream.read(fdri_format.logic_block_size),
                XilinxFdriLogicBlock,
            ),
            DataObject.create_packed(
                self.context,
                data_stream.read(fdri_format.bram_block_size),
                XilinxFdriRAMBlockInterface,
            ),
            DataObject.create_packed(
                self.context,
                data_stream.read(fdri_format.io_block_size),
                XilinxFdriIOBlockInterface
            ),
            DataObject.create_packed(
                self.context,
                data_stream.read(fdri_format.crc_size),
                XilinxFdriCRCInterface,
            ),
        )

    def pack(self, data_object):
        """
        :param XilinxFdriPayload data_object:
        :rtype: bytes
        """
        assert isinstance(data_object, XilinxFdriPayload)
        return b"".join([
            data_object.get_logic_block().pack(),
            data_object.get_ram_block().pack(),
            data_object.get_io_block().pack(),
            data_object.get_crc().pack()
        ])


import ctypes
import json
import struct
from collections import OrderedDict

from typing import List

import six


def hex_to_bytes(hex):
    """
    Convert a hex value to the bytes representation used by the interpreter.

    :param str hex:
    :rtype: bytes
    """
    if six.PY2:
        return hex.decode("hex")
    else:
        return bytes.fromhex(hex)


class XilinxAttributeValueDocumentation:
    """
    Defines the documentation for a specific value of a register attribute.

    :param int value: The attribute's value.
    :param Optional[str] name: The  name for the value of the attribute.
    :param Optional[None] description: The description for the value of the attribute.

    :ivar int value: The attribute's value.
    :ivar Optional[str] name: The  name for the value of the attribute.
    :ivar Optional[None] description: The description for the value of the attribute.
    """
    def __init__(
            self,
            value,
            name,
            description
    ):
        self.value = value
        self.name = name
        self.description = description


class XilinxRegisterAttributeFormat:
    """
    Defines the format/documentation of a register payload attribute.

    :param str name: The name of the attribute.
    :param int bit_size: The size of the attribute value in bits.
    :param str description: A description of the attribute.
    :param List[XilinxAttributeValueDocumentation] values: The documentation for values
        of the attribute.

    :ivar str name: The name of the attribute.
    :ivar int bit_size: The size of the attribute value in bits.
    :ivar str description: A description of the attribute.
    """
    def __init__(
        self,
        name,
        bit_size,
        description,
        values,
    ):
        self.name = name
        self.bit_size = bit_size
        self.description = description
        self._value_config_by_value = {c.value: c for c in values}

    def get_value_documentation(self, value):
        """
        Get the documentation for the provided value of the attribute

        :param int value: The value of the register attribute
        :rtype: Optional[XilinxAttributeValueDocumentation]
        """
        return self._value_config_by_value.get(value)


class XilinxRegisterFormatCtypeLE(ctypes.Union):
    _fields_ = []

    def __init__(self, raw_bytes):
        if len(raw_bytes) == 2:
            struct_flag = ">H"
        elif len(raw_bytes) == 4:
            struct_flag = ">I"
        else:
            raise RuntimeError(
                "XilinxRegisterFormatCtypeLE currently only supports 2 and 4 byte structures"
            )
        self.binary_data = struct.unpack(struct_flag, raw_bytes)[0]


class XilinxRegisterFormatCtype(object):
    def __init__(self, class_name, fields, values=None):
        self.class_name = class_name
        self.fields = fields
        self.values = values
        if six.PY2:
            bit_size = sum([bit_size for _, _, bit_size in fields])
            if bit_size <= 16:
                binary_data_type = ctypes.c_uint16
            elif bit_size == 32:
                binary_data_type = ctypes.c_uint32
            else:
                raise RuntimeError("XilinxRegisterFormatCtype bit size not supported")
            self.bits = type("{}Bits".format(str(class_name)), (ctypes.LittleEndianStructure, ), {
                "_pack_": 1,
                "_fields_": fields[::-1]
            })

            self.ctype = type(str(class_name), (XilinxRegisterFormatCtypeLE, ), {
                "_fields_": [
                    ("bits", self.bits),
                    ("binary_data", binary_data_type)
                ]
            })
        if six.PY3:
            self.ctype = type(str(class_name), (ctypes.BigEndianStructure, ), {
                "_pack_": 1,
                "_fields_": fields
            })

    def get_bytes(self, *values):
        # type: (*int) -> bytes

        if six.PY3:
            ctype_instance = self.ctype(*values)
            return bytearray(bytes(ctype_instance))
        elif six.PY2:
            reversed_values = values[::-1]
            ctype_instance = self.bits(*reversed_values)
            num_packed_bytes = sum([bit_size for _, _, bit_size in self.fields]) / 8
            out_of_order_bytes = bytes(bytearray(ctype_instance))
            bytes_to_reverse = out_of_order_bytes[:num_packed_bytes]
            reversed_bytes = b"".join(
                [byte_to_reverse for byte_to_reverse in bytes_to_reverse][::-1]
            )
            return b"".join([reversed_bytes, out_of_order_bytes[num_packed_bytes:]])
        else:
            raise RuntimeError

    def from_buffer_copy(self, raw_bytes):
        # type: (bytes) -> XilinxRegisterFormatCtype
        if six.PY3:
            ctype_instance = self.ctype.from_buffer_copy(raw_bytes)
            values = []
            for field_name, _, _ in self.fields:
                values.append(ctype_instance.__getattribute__(field_name))
        elif six.PY2:
            num_packed_bytes = sum([bit_size for _, _, bit_size in self.fields]) / 8
            ctype_instance = self.ctype(raw_bytes[:num_packed_bytes])
            values = []
            for field_name, _, _ in self.fields:
                values.append(ctype_instance.bits.__getattribute__(field_name))
        else:
            raise RuntimeError
        return XilinxRegisterFormatCtype(self.class_name, self.fields, values)


class XilinxRegisterFormat:
    """
    Defines the format/documentation for a register packet.

    :param int address: The address of the register.
    :param str name: The name of the register targeted by a packet.
    :param str description: A description of the register.
    :param List[XilinxRegisterAttributeFormat] attributes: The attributes to parse in
        a packet's data.

    :ivar int address: The address of the register.
    :ivar str name: The name of the register targeted by a packet.
    :ivar str description: A description of the register.
    :ivar List[XilinxRegisterAttributeFormat] attributes: The attributes to parse in
        a packet's data.
    :ivar int size: Ths aggregate bit size of the register payload attributes.
    :ivar ctype ctype: A ctype class definition configured to match the register payload
        attributes.
    """
    def __init__(
        self,
        address,
        name,
        description,
        attributes,
    ):
        self.address = address
        self.name = name
        self.description = description
        self.attributes = attributes

        bit_size = sum([attr.bit_size for attr in self.attributes])
        if bit_size % 8 != 0:
            raise ValueError("The bit size of the register config should be a multiple of 8")
        self.size = int(bit_size / 8)

        class_name = "".join([p.title() for p in self.name.split("_")])
        fields = [
            (attr.name.lower(), ctypes.c_uint32, attr.bit_size)
            for attr in self.attributes
        ]
        self.ctype = XilinxRegisterFormatCtype(str(class_name), fields)


class XilinxFdriMajorFormat:
    """
    Defines the format for a specif major type.

    :param str name: The name of the major.
    :param int frame_size: The size of a frame within the major.
    :param int frame_count: The number of frames making up the major.
    :param List[str] frame_descriptions: A description of each frame. The length of the list is
        expected to match the `frame_count` property.

    :ivar str name: The name of the major.
    :ivar int frame_size: The size of a frame within the major.
    :ivar int frame_count: The number of frames making up the major.
    :ivar List[str] frame_descriptions: A description of each frame. The length of the list is
        expected to match the `frame_count` property.
    """
    def __init__(
            self,
            name,
            frame_size,
            frame_count,
            frame_descriptions
    ):
        self.name = name
        self.frame_size = frame_size
        self.frame_count = frame_count
        self.frame_descriptions = frame_descriptions


class XilinxFdriPinFormat:
    """
    Defines the format for a specif io pin.

    :param str name: The name of the pin.
    :param int offset: The offset of the pin.
    :param str on_value: The hex representation of the value to turn the pin on.
    :param str off_value: The hex representation of the value to turn the pin off.

    :ivar str name: The name of the pin.
    :ivar int offset: The offset of the pin.
    :ivar bytes on_value: The hex representation of the value to turn the pin on.
    :ivar bytes off_value: The hex representation of the value to turn the pin off
    """
    def __init__(
            self,
            name,
            offset,
            on_value,
            off_value
    ):
        self.name = name
        self.offset = offset
        self.on_value = hex_to_bytes(on_value)
        self.off_value = hex_to_bytes(off_value)


class XilinxFdriFormat:
    """
    Defines the format of an FDRI register payload for a specific type of FPGA.

    :param str device_name: The name of the FPGA.
    :param int logic_block_size: The size of the logic block in bytes.
    :param int ram_block_size: The size of the ram block in bytes.
    :param int io_block_size: The size of the io block in bytes.
    :param int crc_size: The size of the CRC checksum in bytes.
    :param Optional[List[List[XilinxFdriMajorFormat]]Optional[ logic_block_format: A matrix of Major
        formats that are contained in the logic block..
    :param Optional[List[XilinxFdriPinFormat]] io_block_format: A list of pin formats that are
        contained in the io block.

    :ivar str device_name: The name of the FPGA.
    :ivar int logic_block_size: The size of the logic block in bytes.
    :ivar int ram_block_size: The size of the ram block in bytes.
    :ivar int io_block_size: The size of the io block in bytes.
    :ivar int crc_size: The size of the CRC checksum in bytes.
    :ivar Optional[List[List[XilinxFdriMajorFormat]]Optional[ logic_block_format: A matrix of Major
        formats.
    """
    def __init__(
            self,
            device_name,
            logic_block_size,
            ram_block_size,
            io_block_size,
            crc_size,
            logic_block_format,
            io_block_format,
    ):
        self.device_name = device_name
        self.logic_block_size = logic_block_size
        self.bram_block_size = ram_block_size
        self.io_block_size = io_block_size
        self.crc_size = crc_size
        self.logic_block_format = logic_block_format
        if io_block_format is None:
            self._io_pin_by_name = {}
        else:
            self._io_pin_by_name = {pin.name: pin for pin in io_block_format}

    def get_io_pin_by_name(self, name):
        """
        Retrieve the IO pin format for the provided IO pin name.

        :param str name: The name of the IO pin
        :rtype: XilinxFdriPinFormat|None
        """
        return self._io_pin_by_name.get(name)


class XilinxFormat:
    """
    Defines the format of the registers in a Xilinx bitstream. It contains indexes to look up
    register formats by name or address.

    :param List[XilinxRegisterFormat] registers: The list of register formats.
    :param List[XilinxFdriFormat] fdri_formats: The list of FDRI packets formats targeting
        different types of FPGA.
    :param Any visualizer_style: The style config for the visualizer frontend.
    :param str sync_word: The sync word that marks the beginning of the configuration data in hex
        format.

    :ivar bytes sync_word: The sync word that marks the beginning of the configuration data.
    """
    def __init__(self, registers, fdri_formats, visualizer_style, sync_word="AA995566"):
        self._register_format_by_address = {c.address: c for c in registers}
        self._register_format_by_name = {c.name: c for c in registers}
        self._fdri_format_by_device = {c.device_name: c for c in fdri_formats}
        self.visualizer_style = visualizer_style
        self.sync_word = hex_to_bytes(sync_word)

    def get_fdri_format(self, device_name):
        """
        Lookup the FDRI format for the provided device name
        :param str device_name: The name of the device
        :rtype: XilinxFdriFormat | None
        """
        return self._fdri_format_by_device.get(device_name)

    def get_register_format(self, address):
        """
        Lookup a register format by its address
        :param int address: The register's address
        :rtype: XilinxRegisterFormat | None
        """
        return self._register_format_by_address.get(address)

    def get_register_format_by_name(self, name):
        """
        Lookup a register format by its name
        :param str name: The register's name
        :rtype: XilinxRegisterFormat | None
        """
        return self._register_format_by_name.get(name)


class XilinxFormatBuilder:
    def __init__(self):
        self._visualizer_config = None # type: Any
        self._register_formats = []  # type: List[Any]
        self._fdri_io_block_formats = []  # type: List[Any]
        self._fdri_logic_block_formats = []  # type: List[Any]
        self._fdri_major_formats = []  # type: List[Any]
        self._fdri_formats = []  # type: List[Any]

    def set_visualizer_config(self, visualizer_config):
        self._visualizer_config = visualizer_config

    def add_fdri_logic_block_formats(self, logic_block_formats):
        if not isinstance(logic_block_formats, list):
            raise ValueError("The FDRI logic blocks formats are expected to be a list")
        self._fdri_logic_block_formats.extend(logic_block_formats)

    def add_fdri_io_block_formats(self, io_block_formats):
        if not isinstance(io_block_formats, list):
            raise ValueError("The FDRI IO blocks formats are expected to be a list")
        self._fdri_io_block_formats.extend(io_block_formats)

    def add_register_formats(self, register_formats):
        if not isinstance(register_formats, list):
            raise ValueError("The register formats are expected to be a list")
        self._register_formats.extend(register_formats)

    def add_fdri_formats(self, fdri_formats):
        if not isinstance(fdri_formats, list):
            raise ValueError("The fdri formats are expected to be a list")
        self._fdri_formats.extend(fdri_formats)

    def add_fdri_major_formats(self, major_formats):
        if not isinstance(major_formats, list):
            raise ValueError("The fdri formats are expected to be a list")
        self._fdri_major_formats.extend(major_formats)

    def set_visualizer_config_json(self, path):
        with open(path, "r") as f:
            return self.set_visualizer_config(json.load(f))

    def add_fdri_logic_block_formats_json(self, path):
        with open(path, "r") as f:
            return self.add_fdri_logic_block_formats(json.load(f))

    def add_fdri_io_block_formats_json(self, path):
        with open(path, "r") as f:
            return self.add_fdri_io_block_formats(json.load(f))

    def add_register_formats_json(self, path):
        with open(path, "r") as f:
            return self.add_register_formats(json.load(f))

    def add_fdri_formats_json(self, path):
        with open(path, "r") as f:
            return self.add_fdri_formats(json.load(f))

    def add_fdri_major_formats_json(self, path):
        with open(path, "r") as f:
            return self.add_fdri_major_formats(json.load(f))

    def _create_register_attribute_value_format(self, value_format):
        if not isinstance(value_format, dict):
            raise ValueError("The register attribute value format is expected to be a dict")
        return XilinxAttributeValueDocumentation(
            value_format["value"],
            value_format["name"],
            value_format["description"],
        )

    def _create_register_attribute_format(self, attribute_format):
        if not isinstance(attribute_format, dict):
            raise ValueError("The register attribute format is expected to be a dict")

        return XilinxRegisterAttributeFormat(
            attribute_format["name"],
            attribute_format["bit_size"],
            attribute_format["description"],
            [
                self._create_register_attribute_value_format(value_format)
                for value_format in attribute_format["values"]
            ]
        )

    def _create_register_format(self, register_format):
        if not isinstance(register_format, dict):
            raise ValueError("The register format is expected to be a dict")
        if not isinstance(register_format["attributes"], list):
            raise ValueError("The register format attributes is expected to be a list")
        return XilinxRegisterFormat(
            register_format["address"],
            register_format["name"],
            register_format["description"],
            [
                self._create_register_attribute_format(attribute_format)
                for attribute_format in register_format["attributes"]
            ]
        )

    def _create_fdri_major_format(self, major_format):
        if not isinstance(major_format, dict):
            raise ValueError("The major format is expected to be a dict")
        return XilinxFdriMajorFormat(
            major_format["name"],
            major_format["frame_size"],
            major_format["frame_count"],
            major_format["frame_descriptions"],
        )

    def _create_fdri_logic_block_format(self, logic_block_format, fdri_major_formats_by_name):
        if not isinstance(logic_block_format, list):
            raise ValueError("The logic block format is expected to be a list")
        logic_block_format_obj = []
        for row_format in logic_block_format:
            if not isinstance(row_format, list):
                raise ValueError("The column format is expected to be a list")
            row_format_obj = []
            for major_name in row_format:
                if six.PY2:
                    expected_type = unicode
                elif six.PY3:
                    expected_type = str
                else:
                    raise RuntimeError
                if not isinstance(major_name, expected_type):
                    raise ValueError("The major name should be a string")
                if major_name not in fdri_major_formats_by_name:
                    raise ValueError("No major format named {}".format(major_name))
                row_format_obj.append(fdri_major_formats_by_name[major_name])
            logic_block_format_obj.append(row_format_obj)
        return logic_block_format_obj

    def _create_fdri_io_block_format(
            self,
            fdri_io_block_format,
    ):
        if fdri_io_block_format is None:
            return None
        if not isinstance(fdri_io_block_format, list):
            raise ValueError("The FDRI io block format is expected to be a dict")
        items = []
        for fdri_io_pin_format in fdri_io_block_format:
            items.append(XilinxFdriPinFormat(
                fdri_io_pin_format["pin_name"],
                fdri_io_pin_format["offset"],
                fdri_io_pin_format.get("on_value"),
                fdri_io_pin_format.get("off_value"),
            ))
        return items

    def _create_fdri_format(
            self,
            fdri_format,
            fdri_logic_block_format,
            fdri_io_block_format,
            fdri_major_formats_by_name,
    ):
        if not isinstance(fdri_format, dict):
            raise ValueError("The fdri format is expected to be a dict")

        return XilinxFdriFormat(
            fdri_format["device_name"],
            fdri_format["logic_block_size"],
            fdri_format["bram_block_size"],
            fdri_format["io_block_size"],
            fdri_format["crc_size"],
            self._create_fdri_logic_block_format(
                fdri_logic_block_format,
                fdri_major_formats_by_name
            ),
            self._create_fdri_io_block_format(
                fdri_io_block_format
            ),
        )

    def build(self):
        register_formats = [
            self._create_register_format(register_format)
            for register_format in self._register_formats
        ]
        fdri_major_formats = [
            self._create_fdri_major_format(fdri_format)
            for fdri_format in self._fdri_major_formats
        ]
        fdri_major_formats_by_name = {m.name: m for m in fdri_major_formats}

        fdri_io_block_by_name = {
            m["device_name"]: m["io_block_format"]
            for m in self._fdri_io_block_formats
        }
        fdri_logic_block_by_name = {
            m["device_name"]: m["logic_block_format"]
            for m in self._fdri_logic_block_formats
        }
        fdri_formats = [
            self._create_fdri_format(
                fdri_format,
                fdri_logic_block_by_name.get(fdri_format["device_name"]),
                fdri_io_block_by_name.get(fdri_format["device_name"]),
                fdri_major_formats_by_name,
            )
            for fdri_format in self._fdri_formats
        ]
        return XilinxFormat(
            register_formats,
            fdri_formats,
            self._visualizer_config
        )


from bal_xilinx.analyzers.device_analyzer import XilinxDeviceAnalyzer
from bal_xilinx.context import XilinxContext


class XilinxEncryptionAnalyzer(XilinxDeviceAnalyzer):
    """
    An analyzer used to determine if a Xilinx bitstream it encrypted.

    :param XilinxContext context: The configured xilinx context
    """
    def __init__(self, context):
        super(XilinxDeviceAnalyzer, self).__init__(context)
        self.context = context

    def analyze(self):
        """
        Returns true if the bitstream is encrypted.

        :rtype: bool
        """
        config_packets = self.context.get_data().unpack()\
            .get_packets_by_register_name("Ctl")
        for config_packet in config_packets:
            decryption = config_packet.get_payload().unpack()\
                .get("dec").unpack()
            if decryption.get_value() == 1:
                return True
        return False



import base64

from bal.analyzers.visualizer_analyzer import VisualizerAnalyzer
from bal_xilinx.context import XilinxContext


class XilinxVisualizerAnalyzer(VisualizerAnalyzer):
    """
    Generate nested native objects (ie can be fed to serialization libraries) that are used by
    the visualizer to display the data within the provided data object.

    :param XilinxContext context: The configured xilinx context
    """
    def __init__(self, context):
        super(XilinxVisualizerAnalyzer, self).__init__(context)
        self.context = context

    def analyze(self):
        """
        Build a visualizer config object for the bitstream.
        :rtype: Dict[str, Any]
        """
        bitstream = self.context.get_data()
        bitstream.synchronize()
        bitstream_data = bitstream.pack()
        return {
            "style": self.context.format.visualizer_style,
            "data": self._traverse(bitstream),
            "bytes": base64.b64encode(bitstream_data).decode('ascii')
        }





from bal.context_ioc import AbstractModifier
from bal_xilinx.context import XilinxContext
from bal_xilinx.data_model import XilinxFdriPayload


class XilinxPinModifer(AbstractModifier):
    """
    An analyzer used to modify an FDRI IOB pin config.

    :param XilinxContext context: The configured xilinx context
    """

    def __init__(self, context):
        super(XilinxPinModifer, self).__init__(context)
        self.context = context

    def modify(self, pin_name, on, **kwargs):
        """
        Modify an FDRI IOB pin config.

        :param str pin_name: The name of the pin to modify.
        :param bool on: If True, the pin will be pulled high. Otherwise it will be pulled low.
        :param Any kwargs:
        """
        if self.context.id_code is None:
            raise ValueError(
                "The ID code for the targeted device has not been set on the bitstream context."
            )
        io_pin_format = self.context.format.get_fdri_format(self.context.id_code)\
            .get_io_pin_by_name(pin_name)

        if io_pin_format is None:
            raise ValueError("No format information for the IO pin {}".format(pin_name))

        if on is True:
            io_pin_value = io_pin_format.on_value
        else:
            io_pin_value = io_pin_format.off_value

        if io_pin_value is None:
            raise ValueError("No value configured for pin {}, on={}".format(pin_name, on))

        fdri_packets = self.context.get_data().unpack()\
            .get_packets_by_register_name("Fdri")
        if len(fdri_packets) != 1:
            raise ValueError("A single Fdri register packet is expected in the bitstream")
        fdri_packet = fdri_packets[0]
        fdri_packet_payload = fdri_packet.get_payload().unpack()
        assert isinstance(fdri_packet_payload, XilinxFdriPayload)
        io_block_object = fdri_packet_payload.get_io_block()
        io_block_bytes = io_block_object.get_bytes()
        if io_pin_format.offset + len(io_pin_value) > len(io_block_bytes):
            raise ValueError(
                "Invalid format definition for IO pin {}. The last byte would "
                "be written at offset {} but the IO block data size is only {}".format(
                    pin_name,
                    hex(io_pin_format.offset + len(io_pin_value) - 1),
                    len(io_block_bytes)
                )
            )
        io_block_object.set_bytes(
            io_block_bytes[:io_pin_format.offset] + \
                io_pin_value + \
                io_block_bytes[io_pin_format.offset + len(io_pin_value):]
        )



from bal.context_ioc import AbstractConverter
from bal.data_object import DataObject
from bal_xilinx.converters import WORD_SIZE
from bal_xilinx.data_model import XilinxBitstream, XilinxBitstreamHeaderInterface, \
    XilinxPackets, XilinxBitstreamSyncMarker


class XilinxBitstreamConverter(AbstractConverter):
    """
    Converter for a Xilinx FPGA bitstream

    :param XilinxContext context: A context used to create data objects/converters.
    """

    def __init__(self, context):
        super(XilinxBitstreamConverter, self).__init__(context)
        self.context = context

    def unpack(self, data_bytes):
        """
        :param bytes data_bytes:
        :rtype: XilinxBitstream
        """
        sync_marker = self.context.format.sync_word
        sync_marker_index = data_bytes.find(sync_marker)
        assert sync_marker_index >= 0, \
            "The sync marker is not present in the provided bitstream data"

        assert sync_marker_index + len(sync_marker) < len(data_bytes) - WORD_SIZE, \
            "The configuration data is expected to contain at least one word size worth of data"

        return XilinxBitstream(
            DataObject.create_packed(
                self.context,
                data_bytes[:sync_marker_index],
                XilinxBitstreamHeaderInterface
            ),
            DataObject.create_packed(
                self.context,
                data_bytes[sync_marker_index:sync_marker_index+len(sync_marker)],
                XilinxBitstreamSyncMarker,
            ),
            DataObject.create_packed(
                self.context,
                data_bytes[sync_marker_index + len(sync_marker):],
                XilinxPackets,
            )
        )

    def pack(self, data_model):
        """
        :param XilinxBitstream data_model:
        :rtype: bytes
        """
        assert isinstance(data_model, XilinxBitstream)
        return b"".join([
            data_model.get_header().pack(),
            self.context.format.sync_word,
            data_model.get_packets().pack()
        ])



from bal.data_object import DataObject
from bal.context import BALContextFactory, BALContext
from bal_xilinx.data_model import XilinxBitstream
from bal_xilinx.format import XilinxFormat


class XilinxContext(BALContext):
    """
    See the documentation of :py:class:`~bal.context.BALContext` for an overview of the
    purpose of the context class. In addition, the :py:class:`XilinxContext` stores a reference
    to a :py:class:`~bal_xilinx.format.XilinxFormat` instance that defines the format of a
    Xilinx bitstream.

    :param Dict[Type[ConverterInterface],Type[AbstractConverter]] converters_by_type: The converter
        interfaces mapped to their implementation.
    :param Dict[Type[AnalyzerInterface],Type[AbstractAnalyzer]] analyzers_by_type: The analyzer
        interfaces mapped to their implementation.
    :param Dict[Type[ModifierInterface],Type[AbstractModifier]] modifiers_by_type: The modifier
        interfaces mapped to their implementation.
    :param XilinxFormat bitstream_format: The Xilinx bitstream format configuration.
    :param bytes bytes: The bytes making up the bitstream.
    """
    def __init__(
            self,
            converters_by_type,
            analyzers_by_type,
            modifiers_by_type,
            bitstream_format,
            bytes
    ):
        super(XilinxContext, self).__init__(
            converters_by_type,
            analyzers_by_type,
            modifiers_by_type
        )
        self.id_code = None
        self.format = bitstream_format
        self._bitstream = DataObject.create_packed(self, bytes, XilinxBitstream)

    def get_data(self):
        """
        Get the data object wrapping the bitstream. The returned object starts out packed but may
        be modified by user call.

        :rtype: DataObject[XilinxBitstream]
        """
        return self._bitstream


class XilinxContextFactory(BALContextFactory):
    """
    See the documentation of :py:class:`~bal.context.BALContextFactory` for an overview of the
    purpose of the context factory class. In addition, the :py:class:`XilinxContextFactory` stores
    a reference to a :py:class:`~bal_xilinx.format.XilinxFormat` instance that defines the
    format of a Xilinx bitstream.

    :param XilinxFormat bitstream_format: The Xilinx bitstream format configuration.
    """
    def __init__(self, bitstream_format):
        super(XilinxContextFactory, self).__init__()
        self._format = bitstream_format

    def create(self, data):
        """
        Create an Xilinx FPGA context from the provided bytes.

        :param bytes bytes: The bytes for the Xilinx FPGA bitstream
        :rtype: XilinxContext
        """
        return XilinxContext(
            self._converters_by_type,
            self._analyzers_by_type,
            self._modifiers_by_type,
            self._format,
            data
        )




from abc import ABCMeta
from typing import List

import six

from bal.data_object import DataObject
from bal.data_model import ClassModel, ValueModel, DictModel, ArrayModel, DataModel


class XilinxFdriLogicFrame(DataModel):
    """
    A single frame of the major of the logic block of an FDRI payload.
    """


class XilinxFdriLogicMajor(ArrayModel[DataObject[XilinxFdriLogicFrame]]):
    """
    A single major of the logic block of an FDRI payload.
    """


class XilinxFdriLogicRow(ArrayModel[DataObject[XilinxFdriLogicMajor]]):
    """
    A single row of the logic block of an FDRI payload.
    """


class XilinxFdriLogicBlock(ArrayModel[DataObject[XilinxFdriLogicRow]]):
    """
    The logic block of an FDRI payload.
    """


class XilinxFdriRAMBlockInterface(DataModel):
    """
    The RAM block of an FDRI payload.
    """


class XilinxFdriIOBlockInterface(DataModel):
    """
    The IO block of an FDRI payload.
    """


class XilinxFdriCRCInterface(DataModel):
    """
    The CRC tail of an FDRI payload.
    """


class XilinxType2PayloadInterface(DataModel):
    """
    The payload of a type 2 packet in the Xilinx bitstream.
    """


class XilinxFdriPayload(XilinxType2PayloadInterface, ClassModel[DataObject]):
    """
    The payload of a type 2 FDRI packet in the Xilinx bitstream. It contains configuration for
    logic blocks, ram blocks, and io blocks. It also contains a checksum of the blocks config in
    the tail.
    """

    def __init__(self, logic_block, bram_block, iob_block, crc):
        super(XilinxFdriPayload, self).__init__((
            ("logic_block", self.get_logic_block),
            ("ram_block", self.get_ram_block),
            ("io_block", self.get_io_block),
            ("tail", self.get_crc),
        ))
        self._logic_block = logic_block
        self._bram_block = bram_block
        self._iob_block = iob_block
        self._crc = crc

    def get_logic_block(self):
        """
        Get the data object for the logic block of the FDRI payload.

        :rtype: DataObject[XilinxFdriLogicBlock]
        """
        return self._logic_block

    def get_ram_block(self):
        """
        Get the data object for the RAM block of the FDRI payload.

        :rtype: XilinxFdriRAMBlockInterface
        """
        return self._bram_block

    def get_io_block(self):
        """
        Get the data object for the IO block of the FDRI payload.

        :rtype: XilinxFdriIOBlockInterface
        """
        return self._iob_block

    def get_crc(self):
        """
        Get the data object for the CRC tail of the FDRI payload.
        :rtype: XilinxFdriCRCInterface
        """
        return self._crc

    def set_logic_block(self, logic_block):
        self._synced = False
        self._logic_block = logic_block
        return self

    def set_ram_block(self, bram_block):
        self._synced = False
        self._bram_block = bram_block
        return self

    def set_io_block(self, iob_block):
        self._synced = False
        self._iob_block = iob_block
        return self

    def set_tail(self, tail):
        self._synced = False
        self._crc = tail
        return self


class XilinxType1PayloadAttribute(ValueModel):
    """
    An attibute of the payload of a type 1 packet in the Xilinx bitstream. It is parsed
    automatically using the Xilinx format configuration.
    """


class XilinxType1Payload(DictModel[DataObject[XilinxType1PayloadAttribute]]):
    """
    The payload of a type 1 packet in the Xilinx bitstream. It is parsed automatically from using
    the Xilinx format configuration. It's properties are not known until it is unpacked.
    """


class XilinxPacketHeader(ClassModel[DataObject]):
    """
    The header of a Xilinx register configuration packet. It contains the type of the packet (1
    or 2), the opcode (READ/WRITE/NOOP), the register address and the number of words in the
    payload.
    """

    def __init__(self, packet_type, opcode, register_address, word_count):
        super(XilinxPacketHeader, self).__init__((
            ("type", self.get_packet_type),
            ("opcode", self.get_opcode),
            ("register_address", self.get_register_address),
            ("word_count", self.get_word_count),
        ))
        self._type = packet_type
        self._opcode = opcode
        self._register_address = register_address
        self._word_count = word_count

    def get_packet_type(self):
        """
        :rtype: DataObject[ValueModel]
        """
        return self._type

    def get_opcode(self):
        """
        :rtype: DataObject[ValueModel]
        """
        return self._opcode

    def get_register_address(self):
        """
        :rtype: DataObject[ValueModel]
        """
        return self._register_address

    def get_word_count(self):
        """
        :rtype: DataObject[ValueModel]
        """
        return self._word_count

    def set_packet_type(self, packet_type):
        self._synced = False
        self._type = packet_type
        return self

    def set_opcode(self, opcode):
        self._synced = False
        self._opcode = opcode
        return self

    def set_register_address(self, register_address):
        self._synced = False
        self._register_address = register_address
        return self

    def set_word_count(self, word_count):
        self._synced = False
        self._word_count = word_count
        return self


class XilinxPacket(ClassModel[DataObject]):
    """
    A packet within a Xilinx bitstream. It contains a header and a payload.
    """

    def __init__(
            self,
            header,
            payload_size,
            payload,
    ):
        self._header = header
        self._payload_size = payload_size
        self._payload = payload
        super(XilinxPacket, self).__init__((
            ("header", self.get_header),
            ("payload_size", self.get_payload_size),
            ("payload", self.get_payload)
        ))

    def get_header(self):
        """
        :rtype: DataObject[XilinxPacketHeader]
        """
        return self._header

    def get_payload_size(self):
        """
        :rtype: DataObject[ValueModel]
        """
        return self._payload_size

    def get_payload(self):
        """
        :rtype: DataObject
        """
        return self._payload

    def set_header(self, header):
        self._synced = False
        self._header = header
        return self

    def set_payload(self, payload):
        self._synced = False
        self._payload = payload
        return self


class XilinxPacketsTail(DataModel):
    """
    The tail data for the packets. It cannot be detected by the XilinxBitstreamConverter so it
    must be handled as any other packet.
    """
    pass


class XilinxPackets(ArrayModel[DataObject[XilinxPacket]]):
    """
    An array of Xilinx register configuration packet.
    """

    def __init__(self, items):
        super(XilinxPackets, self).__init__(items)


class XilinxBitstreamHeaderInterface(DataModel):
    """
    The Xilinx bitstream header contains unknown information.
    """


class  XilinxBitstreamSyncMarker(DataModel):
    """
    The Xilinx bitstream sync marker
    """


class XilinxBitstream(ClassModel[DataObject]):
    """
    The root model for a Xilinx bitstream. It contains a header and packets data objects.
    """

    def __init__(
            self,
            header,
            sync_marker,
            packets
    ):
        """
        :param DataObject[XilinxBitstreamHeaderInterface] header:
        :param DataObject[XilinxBitstreamSyncMarker] sync_marker:
        :param DataObject[XilinxPackets] packets:
        """
        super(XilinxBitstream, self).__init__((
            ("header", self.get_header),
            ("sync_marker", self.get_sync_marker),
            ("packets", self.get_packets),
        ))
        self._header = header
        self._sync_marker = sync_marker
        self._packets = packets
        self._packets_by_register = {}
        for packet_object in self._packets.unpack():
            if not packet_object.is_unpacked() and not packet_object.is_convertible():
                continue
            packet = packet_object.unpack()
            register_name = packet.get_header().unpack()\
                .get_register_address().unpack()\
                .value_name
            register_packets = self._packets_by_register.get(register_name)
            if register_packets is None:
                register_packets = []
                self._packets_by_register[register_name] = register_packets
            register_packets.append(packet)

    def get_header(self):
        """
        :rtype: DataObject[XilinxBitstreamHeaderInterface]
        """
        return self._header

    def get_sync_marker(self):
        """
        :rtype: DataObject[XilinxBitstreamSyncMarker]
        """
        return self._sync_marker

    def get_packets(self):
        """
        :rtype: DataObject[XilinxPackets]
        """
        return self._packets

    def get_packets_by_register_name(self, register):
        """
        :param str register:
        :rtype: List[XilinxPacket]
        """
        return self._packets_by_register.get(register) or tuple()

    def set_header(self, header):
        """
        :param DataObject header:
        """
        self._synced = False
        self._header = header
        return self

    def set_packets(self, packets):
        self._synced = False
        self._packets = packets
        return self



from bal.context_ioc import AbstractAnalyzer
from bal_xilinx.context import XilinxContext


class XilinxDeviceAnalyzer(AbstractAnalyzer):
    """
    An analyzer used to retrieve the device type from a Xilinx bitstream.

    :param XilinxContext context: The configured xilinx context
    """
    def __init__(self, context):
        super(XilinxDeviceAnalyzer, self).__init__(context)
        self.context = context

    def analyze(self, **kwargs):
        """
        Returns the device name and it's int representation in the bitstream
        :rtype: str
        """
        if self.context.id_code is not None:
            return self.context.id_code
        config_packets = self.context.get_data().unpack()\
            .get_packets_by_register_name("Idcode")
        if len(config_packets) != 1:
            raise ValueError("A single Idcode register packet is expected in the bitstream")
        config_packet = config_packets[0]
        config_payload_object = config_packet.get_payload()
        config_payload = config_payload_object.unpack()
        id_code_value = config_payload.get("idcode").unpack()
        self.context.id_code = id_code_value.value_name
        return self.context.id_code





from setuptools import setup

setup(
    name='bal-xilinx',
    version='0.1',
    description='An implementation of the BAL framework for Xilinx FPGAs',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    author='Red Balloon Security',
    author_email='quack-tech@redballoonsecurity.com',
    packages=[
        'bal_xilinx',
        'bal_xilinx.analyzers',
        'bal_xilinx.converters',
        'bal_xilinx.modifiers',
        'bal_xilinx.tools'
    ],
    include_package_data=True,
    install_requires=[
        'bal',
        'six',
        'typing;python_version<"3.5"'
    ],
    extras_requires={
        'docs': ['sphinx', 'sphinx-rtd', 'm2r'],
        'examples': ['wget'],
        'tests': ['pytest']
    },
)



import os

from bal.analyzers.visualizer_analyzer import VisualizerAnalyzer
from bal_xilinx.analyzers.device_analyzer import XilinxDeviceAnalyzer
from bal_xilinx.analyzers.encryption_analyzer import XilinxEncryptionAnalyzer
from bal_xilinx.analyzers.visualizer_analyzer import XilinxVisualizerAnalyzer
from bal_xilinx.converters.bitstream import XilinxBitstreamConverter
from bal_xilinx.converters.bitstream_packets import XilinxPacketsConverter
from bal_xilinx.converters.bitstream_packets_1 import XilinxType1PayloadConverter
from bal_xilinx.converters.bitstream_packets_fdri import XilinxFdriPayloadConverter, \
    XilinxFdriLogicConverter, XilinxFdriLogicBlockRowConverter, XilinxFdriLogicMajorConverter
from bal_xilinx.data_model import XilinxBitstream, XilinxPackets, \
    XilinxType1Payload, XilinxFdriPayload, XilinxFdriLogicBlock, \
    XilinxFdriLogicRow, XilinxFdriLogicMajor
from bal_xilinx.format import XilinxFormatBuilder
from bal_xilinx.modifiers.pin_modifier import XilinxPinModifer


def register_defaults_context_converters(context):
    context.register_converter(XilinxBitstream, XilinxBitstreamConverter)
    context.register_converter(XilinxPackets, XilinxPacketsConverter)
    context.register_converter(XilinxType1Payload, XilinxType1PayloadConverter)
    context.register_converter(XilinxFdriPayload, XilinxFdriPayloadConverter)
    context.register_converter(XilinxFdriLogicBlock, XilinxFdriLogicConverter)
    context.register_converter(XilinxFdriLogicRow, XilinxFdriLogicBlockRowConverter)
    context.register_converter(XilinxFdriLogicMajor, XilinxFdriLogicMajorConverter)
    return context


def register_defaults_context_analyzers(context):
    context.register_analyzer(VisualizerAnalyzer, XilinxVisualizerAnalyzer)
    context.register_analyzer(XilinxDeviceAnalyzer, XilinxDeviceAnalyzer)
    context.register_analyzer(XilinxEncryptionAnalyzer, XilinxEncryptionAnalyzer)
    return context


def register_defaults_context_modifiers(context):
    context.register_modifier(XilinxPinModifer, XilinxPinModifer)
    return context


def default_xilinx_context(context):
    register_defaults_context_converters(context)
    register_defaults_context_analyzers(context)
    register_defaults_context_modifiers(context)

    return context


def default_xilinx_formats(format_builder):
    """
    Load the default JSON config files for Xilinx fpgas.

    :rtype: XilinxFormatBuilder
    """
    root_path = os.path.join(os.path.dirname(__file__), "configs")
    # Configure the register formats
    format_builder.add_register_formats_json(os.path.join(root_path, "xilinx_registers.json"))
    # Configure the FDRI major formats
    format_builder.add_fdri_major_formats_json(os.path.join(root_path, "xilinx_fdri_majors.json"))
    # Configure the style
    format_builder.set_visualizer_config_json(os.path.join(root_path, "xilinx_visualization.json"))
    # Configure the FDRI formats
    # LX9
    format_builder.add_fdri_formats_json(os.path.join(root_path, "lx9_fdri.json"))
    format_builder.add_fdri_logic_block_formats_json(
        os.path.join(root_path, "lx9_fdri_logic_block.json")
    )
    format_builder.add_fdri_io_block_formats_json(
        os.path.join(root_path, "lx9_fdri_io_block.json")
    )
    # LX45t
    format_builder.add_fdri_formats_json(os.path.join(root_path, "lx45t_fdri.json"))
    format_builder.add_fdri_logic_block_formats_json(
        os.path.join(root_path, "lx45t_fdri_logic_block.json")
    )
    return format_builder



import ctypes
import io
import struct

import six
from typing import Tuple

from bal.context_ioc import AbstractConverter
from bal.data_model import ValueModel
from bal.data_object import DataObject
from bal_xilinx.converters import WORD_SIZE
from bal_xilinx.data_model import XilinxPacket, XilinxPackets, \
    XilinxPacketHeader, XilinxType1Payload, XilinxType2PayloadInterface, \
    XilinxFdriPayload, XilinxPacketsTail
from bal_xilinx.format import XilinxRegisterFormat


class XilinxCtypePacketBigEndianHeaderBase(ctypes.BigEndianStructure):
    word_count = 0
    register_address = 0
    opcode = 0
    type = 0

    _pack_ = 1
    _fields_ = [
        # In a type 2 packet word_count is unused
        ("type", ctypes.c_uint32, 3),
        ("opcode", ctypes.c_uint32, 2),
        ("register_address", ctypes.c_uint32, 6),
        ("word_count", ctypes.c_uint32, 5),
    ]


class XilinxCtypeLittleEndianPacketHeaderBits(ctypes.LittleEndianStructure):
    word_count = 0
    register_address = 0
    opcode = 0
    type = 0

    _pack_ = 1
    _fields_ = [
        # In a type 2 packet word_count is unused
        ("word_count", ctypes.c_uint32, 5),
        ("register_address", ctypes.c_uint32, 6),
        ("opcode", ctypes.c_uint32, 2),
        ("type", ctypes.c_uint32, 3),
    ]


class XilinxCtypeLittleEndianPacketHeader(ctypes.Union):
    _fields_ = [
        ("bits", XilinxCtypeLittleEndianPacketHeaderBits),
        ("binary_data", ctypes.c_uint16)
    ]

    def __init__(self, val):
        self.binary_data = struct.unpack(">H", val)[0]


class XilinxCtypePacketHeader(object):
    def __init__(self, type=0, opcode=0, register_address=0, word_count=0):
        self.type = type
        self.opcode = opcode
        self.register_address = register_address
        self.word_count = word_count

    def get_bytes(self):
        # type: () -> bytes
        if six.PY3:
            return \
                bytes(
                    bytearray(
                        XilinxCtypePacketBigEndianHeaderBase(
                            self.type,
                            self.opcode,
                            self.register_address,
                            self.word_count
                        )
                    )
                )
        elif six.PY2:
            header = XilinxCtypeLittleEndianPacketHeaderBits(
                self.word_count,
                self.register_address,
                self.opcode,
                self.type
            )
            out_of_order_bytes = bytes(bytearray(header))
            return b"".join([out_of_order_bytes[1], out_of_order_bytes[0], out_of_order_bytes[2:]])
        else:
            raise NotImplementedError

    @staticmethod
    def from_buffer_copy(raw_bytes):
        # type: (bytes) -> XilinxCtypePacketHeader
        if six.PY3:
            header = XilinxCtypePacketBigEndianHeaderBase.from_buffer_copy(raw_bytes)
            return XilinxCtypePacketHeader(
                header.type,
                header.opcode,
                header.register_address,
                header.word_count
            )
        elif six.PY2:
            header = XilinxCtypeLittleEndianPacketHeader(raw_bytes[:2])
            return XilinxCtypePacketHeader(
                header.bits.type,
                header.bits.opcode,
                header.bits.register_address,
                header.bits.word_count
            )
        else:
            raise NotImplementedError


class TypeValue(ValueModel):
    pass


class OpCodeValue(ValueModel):
    pass


class RegisterAddressValue(ValueModel):
    pass


class WordCountValue(ValueModel):
    pass


class XilinxPacketsConverter(AbstractConverter):
    """
    Unpacker for a Xilinx FPGA bitstream

    :param XilinxContext context: A factory used to create data objects/converters.
    :ivar XilinxContext context: A factory used to create data objects/converters.
    """
    def __init__(self, context):
        super(XilinxPacketsConverter, self).__init__(context)
        self.context = context

    def _get_register_format(self, header):
        """
        :param XilinxCtypePacketHeader header:
        :rtype: XilinxRegisterFormat
        """
        register_format = self.context.format.get_register_format(header.register_address)
        # Corner case for a the FAR_MAJ register
        if register_format is not None and register_format.name == "FarMaj" and \
                header.word_count > 1:
            register_format = self.context.format.get_register_format_by_name("FarMajExtended")
        # Verify that we know how to parse the config for the register
        if register_format is None:
            raise ValueError(
                "No register format found for address {}".format(hex(header.register_address))
            )
        return register_format

    def _create_type1_payload(self, data_stream, header, register_format):
        """
        Unpack the payload for a type 1 packet

        :param io.BytesIO data_stream:
        :param XilinxCtypePacketHeader header:
        :param XilinxRegisterFormat register_format:
        :rtype: Tuple[DataObject, bool]
        """
        if header.word_count == 0:
            return None, False
        payload_data = data_stream.read(header.word_count * WORD_SIZE)
        if header.opcode == 0:
            raise ValueError("NOOP Xilinx packets are expected to have a 0 length payload")

        payload = DataObject.create_packed(
            self.context,
            payload_data,
            XilinxType1Payload,
            converter_args=(register_format, ),
        )
        if register_format.name == "Cmd":
            return payload, payload.unpack().get("command").get_model().value_name == "DESYNC"

        return payload, False

    def _create_type2_payload(self, data_stream, header):
        """
        Unpack the payload for a type 2 packet
        :param io.BytesIO data_stream:
        :param XilinxCtypePacketHeader header:
        :rtype: Tuple[DataObject,DataObject]
        """
        register_format = self._get_register_format(header)
        word_count_data = data_stream.read(4)
        word_count, = struct.unpack(">I", word_count_data)
        payload_size_object = DataObject.create_unpacked(
            self.context,
            WordCountValue(word_count),
            bytes=word_count_data
        )
        payload_data = data_stream.read((word_count + 2) * WORD_SIZE)
        # TODO: Remove this hack for the LX45T, it's got 2 FDRI packets, one is waaaaay too small.
        if register_format.name != "Fdri" or len(payload_data) < 500:
            return payload_size_object, DataObject.create_packed(
                self.context,
                payload_data,
                XilinxType2PayloadInterface,
                converter_args=(register_format,)
            )

        return payload_size_object, DataObject.create_packed(
            self.context,
            payload_data,
            XilinxFdriPayload
        )

    def _get_type_name(self, type):
        """
        Get the name for the type value
        :param int type:
        :rtype: str
        """
        if type == 0:
            return "NOOP"
        elif type == 1:
            return "Type1"
        elif type == 2:
            return "Type2"
        else:
            raise ValueError(
                "Unexpected header type {} while parsing the Xilinx config packets".format(type)
            )

    def _get_opcode_name(self, opcode):
        """
        Get the name for the opcode value
        :param int opcode:
        :rtype: str
        """
        if opcode == 0:
            return "NOOP"
        elif opcode == 1:
            return "READ"
        elif opcode == 2:
            return "WRITE"
        else:
            raise ValueError(
                "Unexpected header opcode {} while parsing the Xilinx config packets".format(opcode)
            )

    def unpack(self, data_bytes):
        """
        :param bytes data_bytes:
        :rtype: XilinxPackets
        """
        packets = []
        data_stream = io.BytesIO(data_bytes)
        previous_packet_type = 0
        while data_stream.tell() < len(data_bytes):
            header_raw_data = data_stream.read(2)
            header = XilinxCtypePacketHeader.from_buffer_copy(
                bytearray([
                    header_raw_data[0],
                    header_raw_data[1],
                    0,
                    0
                ])
            )
            is_done = False
            register_format = self._get_register_format(header)
            if header.type == 0:
                payload_object = None
                payload_size_object = None
            elif header.type == 1:
                payload_object, is_done = self._create_type1_payload(
                    data_stream,
                    header,
                    register_format
                )
                payload_size_object = None
            elif header.type == 2:
                if previous_packet_type != 1:
                    raise ValueError(
                        "Unexpected packet type 2 after a packet type {}".format(
                            previous_packet_type
                        )
                    )
                payload_size_object, payload_object = self._create_type2_payload(
                    data_stream,
                    header
                )
            else:
                raise ValueError(
                    "Unexpected packet type {} while parsing the Xilinx bitstream".format(
                        header.type
                    )
                )
            previous_packet_type = header.type
            raw_packet_data = header_raw_data
            if payload_size_object is not None:
                raw_packet_data += payload_size_object.get_bytes()
            if payload_object is not None:
                raw_packet_data += payload_object.get_bytes()

            header_object = DataObject.create_unpacked(
                self.context,
                model=XilinxPacketHeader(
                    DataObject.create_unpacked(
                        self.context,
                        TypeValue(int(header.type), self._get_type_name(header.type)),
                        3,
                    ),
                    DataObject.create_unpacked(
                        self.context,
                        OpCodeValue(int(header.opcode), self._get_opcode_name(header.opcode)),
                        2,
                    ),
                    DataObject.create_unpacked(
                        self.context,
                        RegisterAddressValue(int(header.register_address), register_format.name),
                        6,
                    ),
                    DataObject.create_unpacked(
                        self.context,
                        WordCountValue(int(header.word_count)),
                        5,
                    ),
                ),
                bytes=header_raw_data,
            )
            packets.append(DataObject.create_unpacked(
                self.context,
                model=XilinxPacket(
                    header_object,
                    payload_size_object,
                    payload_object,
                ),
                bytes=raw_packet_data,
            ))
            if is_done is True:
                packets.append(DataObject.create_packed(
                    self.context,
                    data_stream.read(),
                    XilinxPacketsTail,
                ))
                break
        return XilinxPackets(packets)

    def pack(self, packets):
        """
        :param XilinxPackets packets:
        :rtype: bytes
        """
        assert isinstance(packets, XilinxPackets)
        parts = []
        for packet_object in packets:
            if not packet_object.is_unpacked():
                parts.append(packet_object.get_bytes())
                continue
            packet = packet_object.get_model()
            header = packet.get_header().get_model()
            packet_type = header.get_packet_type().get_model().get_value()
            register_address = header.get_register_address().get_model().get_value()
            opcode = header.get_opcode().get_model().get_value()

            payload = packet.get_payload()
            payload_raw = payload.pack() if payload is not None else b""
            if packet_type == 0 or packet_type == 1:
                header = XilinxCtypePacketHeader(
                    packet_type,
                    opcode,
                    register_address,
                    int(len(payload_raw) / WORD_SIZE),
                )
                header_raw = header.get_bytes()
                parts.append(header_raw[:2])
                if opcode != 0:
                    register_format = self._get_register_format(header)
                    assert register_format.size == len(payload_raw), \
                        "The payload size ({}) does not match the expected payload size ({}) for " \
                        "register {}".format(
                            len(payload_raw),
                            register_format.size,
                            register_format.name
                        )
                    parts.append(payload_raw)
            elif packet_type == 2:
                header = XilinxCtypePacketHeader(
                    packet_type,
                    opcode,
                    register_address,
                    0,
                )
                header_raw = header.get_bytes()
                parts.append(header_raw[:2])
                parts.append(struct.pack(">I", int(len(payload_raw) / WORD_SIZE) - 2))
                parts.append(payload_raw)
        return b"".join(parts)


